
unit MainUnit;

interface

uses
   Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
   DDraw, ExtCtrls, Module1, StdCtrls;

const
   MAX_IMAGE_SURFACE = 1;
type
   TForm1 = class(TForm)
      Label1  : TLabel;
      procedure FormCreate(Sender : TObject);
      procedure FormDestroy(Sender : TObject);
      procedure FormShow(Sender : TObject);
      procedure FormKeyDown(Sender : TObject; var Key: Word; Shift : TShiftState);
      procedure FormPaint(Sender : TObject);

   private
      FFlippingEnabled : boolean;

      procedure ExceptionHandler(Sender : TObject; E : Exception);
      function  RestoreSurfaces : HResult;
      procedure DrawSurfaces;
      function  GetBitmapName(i : integer) : string;
      procedure HandleMessage(var Msg : TMsg; var Handled : boolean);
      procedure IdleHandler(Sender : TObject; var Done : boolean);
      procedure SetFlippingEnabled(Value : boolean);
      procedure UpdateDisplay;

   public
      DirectDraw       : IDirectDraw;
      PrimarySurface   : IDirectDrawSurface;
      BackBuffer       : IDirectDrawSurface;
      BackCanvas       : TCanvas;
      DDPalette        : IDirectDrawPalette;
      Image            : array[1..MAX_IMAGE_SURFACE] of IDirectDrawSurface;

      function  MakeItSo(DDResult : HResult) : boolean;
      property  FlippingEnabled : boolean read FFlippingEnabled
                                          write SetFlippingEnabled;
   end;

var
   Form1: TForm1;

   StartTickCount : integer = 0;
   FlippingCount  : integer = -20;

implementation

{$R *.DFM}

uses
   DDUtils;

const
   BitmapName = 'NeTo';


procedure TForm1.FormCreate(Sender: TObject);
begin
   if DirectDrawCreate(nil,DirectDraw,nil) <> DD_OK then
      Raise Exception.Create('DirectDraw ü  ');

   Application.OnException := ExceptionHandler;

   BorderStyle := bsNone;
   Color       := clBlack;
   Cursor      := crNone;
end;

procedure TForm1.FormDestroy(Sender: TObject);
var
   i : integer;
begin
   FlippingEnabled := FALSE;

   BackCanvas.Free;

   if Assigned(PrimarySurface) then PrimarySurface.Release;
   for i := 1 to MAX_IMAGE_SURFACE do begin
      if Assigned(Image[i]) then Image[i].Release;
   end;

   if Assigned(DDPalette) then DDPalette.Release;

   if Assigned(DirectDraw) then DirectDraw.Release;
   DirectDraw := nil;

   Application.OnException := nil;
end;

procedure TForm1.FormShow(Sender: TObject);
var
   i              : integer;
   ADDSurfaceDesc : TDDSurfaceDesc;
   ADDSCaps       : TDDSCaps;
   AColorKey      : TDDColorKey;
begin
   if not Assigned(DirectDraw) then exit;

   if DirectDraw.SetCooperativeLevel(Handle,DDSCL_EXCLUSIVE or DDSCL_FULLSCREEN) <> DD_OK then
      Raise Exception.Create('Ÿ Ǯũ  ');

   if DirectDraw.SetDisplayMode(640,480,8) <> DD_OK then
      Raise Exception.Create('640 x 480 x 8bit  ');

   FillChar(ADDSurfaceDesc,SizeOf(ADDSurfaceDesc),0);
   with ADDSurfaceDesc do begin
      dwSize            := SizeOf(ADDSurfaceDesc);
      dwFlags           := DDSD_CAPS or DDSD_BACKBUFFERCOUNT;
      ddSCaps.dwCaps    := DDSCAPS_COMPLEX or DDSCAPS_FLIP or DDSCAPS_PRIMARYSURFACE;
      dwBackBufferCount := 1;
   end;

   if DirectDraw.CreateSurface(ADDSurfaceDesc,PrimarySurface,nil) <> DD_OK then
      Raise Exception.Create(' ȭ   ');

   ADDSCaps.dwCaps := DDSCAPS_BACKBUFFER;
   if PrimarySurface.GetAttachedSurface(ADDSCaps,BackBuffer) <> DD_OK then
      Raise Exception.Create('ĸ   ');

   BackCanvas := TCanvas.Create;

   DDPalette := DDLoadPalette(DirectDraw,GetBitmapName(1));
   if PrimarySurface.SetPalette(DDPalette) <> DD_OK then
      Raise Exception.Create('ȷƮ ε ');

   for i := 1 to MAX_IMAGE_SURFACE do begin
      Image[i] := DDLoadBitmap(DirectDraw,GetBitmapName(i),0,0);
   end;

   for i := 1 to MAX_IMAGE_SURFACE do begin
      AColorKey.dwColorSpaceLowValue := 0;
      AColorKey.dwColorSpaceHighValue := 0;
      if Image[i].SetColorKey(DDCKEY_SRCBLT,AColorKey) <> DD_OK then
         Raise Exception.Create('  ');
   end;

   initializeGame;

   FlippingEnabled := TRUE;
end;

procedure TForm1.ExceptionHandler(Sender : TObject; E : Exception);
begin
   FlippingEnabled := FALSE;
   if Assigned(DirectDraw) then DirectDraw.FlipToGDISurface;
   MessageDlg(E.Message,mtError,[ mbOK ],0);
   Close;
end;

function TForm1.MakeItSo(DDResult : HResult) : boolean;
begin
   case DDResult of
      DD_OK             : Result := TRUE;
      DDERR_SURFACELOST : Result := RestoreSurfaces <> DD_OK;
      else                Result := DDResult <> DDERR_WASSTILLDRAWING;
   end;
end;

function TForm1.RestoreSurfaces : HResult;
var
   i : integer;
begin
   Result := PrimarySurface.Restore;
   if Result = DD_OK then begin
      for i := 1 to MAX_IMAGE_SURFACE do begin
         Result := Image[i].Restore;
         if Result <> DD_OK then begin
            exit;
         end;
      end;
      if Result = DD_OK then DrawSurfaces;
   end;
end;

procedure TForm1.DrawSurfaces;
var
   i : integer;
begin
   for i := 1 to MAX_IMAGE_SURFACE do begin
      DDReloadBitmap(Image[i],GetBitmapName(i));
   end;
end;

function TForm1.GetBitmapName(i : integer) : string;
begin
   Result := ExtractFilePath(Application.ExeName) + BitmapName + IntToStr(i) + '.BMP';
end;

procedure TForm1.HandleMessage(var Msg : TMsg; var Handled : boolean);
begin
   UpdateDisplay;
end;

procedure TForm1.IdleHandler(Sender : TObject; var Done : boolean);
begin
   UpdateDisplay;
   Done := FALSE;
end;

procedure TForm1.SetFlippingEnabled(Value : boolean);
begin
   if Value <> FFlippingEnabled then begin
      FFlippingEnabled := Value;
      if FFlippingEnabled then begin
         Application.OnMessage := HandleMessage;
         Application.OnIdle    := IdleHandler;
      end else begin
         Application.OnMessage := nil;
         Application.OnIdle    := nil;
      end;
   end;
end;

procedure TForm1.FormPaint(Sender: TObject);
begin
   DrawSurfaces;
end;

procedure TForm1.FormKeyDown(Sender: TObject; var Key: Word; Shift : TShiftState);
begin
   if key = VK_ESCAPE then Close;
   if key = VK_F12 then Close;
   if KeyBuffer.is_installed then KeyBuffer.Key := key;
end;

procedure TForm1.UpdateDisplay;
var
   i  : integer;
   s  : string;
   DC : HDC;
begin
   if FlippingCount = 0 then begin
      StartTickCount := GetTickCount;
   end;
   inc(FlippingCount);

   drawScroll;

   for i := 1 to MAX_PERSON do begin
      person[i].doAction;
   end;

   if FlippingCount > 0 then begin
      if BackBuffer.GetDC(DC) = DD_OK then begin
         with BackCanvas do try
            Handle := DC;
            Font.Name := 'Arial';
            Font.Size := 12;
            Brush.Color := clBlue;
            Font.Color := clWhite;
            str(((FlippingCount*1000) / (GetTickCount-StartTickCount)) : 5 : 2,s);
            Textout(50,50,'FPS : '+s);
         finally
            Handle := 0;
            BackBuffer.ReleaseDC(DC);
         end;
      end;
   end;

   repeat
   until MakeItSo(PrimarySurface.Flip(nil,DDFLIP_WAIT));
end;

end.
