unit Module1;

interface

const
   MAX_X_LINE    = 640;
   MAX_Y_LINE    = 480;
   MAX_MAP_X     = 100;
   MAX_MAP_Y     = 100;
   MAX_PERSON    = 3;
   PERSON_FRAME  = 8;

   MAX_KEY_BUFFER = 32;

type
   TKeyBuffer = object

      is_installed        : boolean;
      constructor init;
      destructor  done;

   private
      top_key, bottom_key : integer;
      KeyBuffer : array[1..MAX_KEY_BUFFER] of word;

      function    readKey : word;
      procedure   writeKey(data : word);

   public
      function    pressedKey : boolean;
      procedure   clearBuffer;
      property    Key : word read readKey write writeKey;
   end;

   TMapAttribute = (Emap_tile,Emap_height,Emap_person);
   TMapData      = array[0..pred(MAX_MAP_Y),0..pred(MAX_MAP_X),TMapAttribute] of byte;

   TMap = object

      constructor init;
      destructor  done;

   private
      m_map_data : TMapData;

      function    readMap(attribute : TMapAttribute; x, y : integer) : byte;
      procedure   writeMap(attribute : TMapAttribute; x, y : integer; value : byte);

   public
      function    isMovableMap(_x, _y, x, y : integer) : boolean;
      property    Map[attribute : TMapAttribute; x, y : integer] : byte read readMap write writeMap;
   end;

   TPerson = object

      m_number, m_x, m_y  : integer;
      m_face, m_face_count : integer;

      constructor init(number : integer);
      destructor  done;
      procedure   setFace(face : integer);
      function    returnCharaTile : integer;
      function    returnX : integer;
      function    returnY : integer;
      function    returnFace : integer;
      procedure   moveXY(x1, y1 : integer);
      procedure   warpXY(x, y : integer);
      procedure   doAction;
   end;

var
   KeyBuffer : TKeyBuffer;
   map       : TMap;
   person    : array[1..MAX_PERSON] of TPerson;
   MAIN_CHARACTER : integer = 1;


procedure initializeGame;
procedure drawScroll;

implementation
uses
   Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
   DDraw, ExtCtrls, MainUnit;

const
   MAP_X_WIDE    = 4;
   MAP_Y_WIDE    = 5;
   TILE_X_SIZE   = 60;
   TILE_X_HALF   = TILE_X_SIZE div 2;
   TILE_Y_SIZE   = 51;
   TILE_Y_HALF   = 14;

   MAP_SQ_WIDE   = 20;
   MAP_X_CENTER  = (MAX_X_LINE - TILE_X_SIZE) div 2 + 310;
   MAP_Y_CENTER  = (MAX_Y_LINE - TILE_Y_SIZE) div 2 + 80;

var
   tile  : array[0..4*6] of TRect;
   chara : array[0..8*MAX_PERSON] of TRect;


constructor TKeyBuffer.init;
begin
   top_key := 1;
   bottom_key := 1;
   is_installed := TRUE;
end;

destructor  TKeyBuffer.done;
begin
   is_installed := FALSE;
end;

function    TKeyBuffer.pressedKey : boolean;
begin
   pressedKey := top_key <> bottom_key;
end;

procedure   TKeyBuffer.clearBuffer;
begin
   top_key := 1;
   bottom_key := 1;
end;

function    TKeyBuffer.readKey : word;
begin
   if pressedKey then begin
      readKey := KeyBuffer[bottom_key];
      inc(bottom_key);
      if bottom_key > MAX_KEY_BUFFER then begin
         bottom_key := 1;
      end;
   end else readKey := 0;
end;

procedure   TKeyBuffer.writeKey(data : word);
begin
   KeyBuffer[top_key] := data;
   inc(top_key);
   if top_key > MAX_KEY_BUFFER then begin
      top_key := 1;
   end;
   if top_key = bottom_key then begin
      inc(bottom_key);
      if bottom_key > MAX_KEY_BUFFER then begin
         bottom_key := 1;
      end;
   end;
end;

(* ........... *)

constructor TMap.init;
var
   i, j, k : integer;
   f       : file;
begin

   AssignFile(f,ExtractFilePath(Application.ExeName) + 'Map.Dat');
  {$I-}
   reset(f,1);
  {$I+}
   if IOResult = 0 then begin
      BlockRead(f,m_map_data[0,0,Emap_tile],sizeof(m_map_data));
      CloseFile(f);

      for k := 1 to MAX_PERSON do
      for j := 0 to pred(MAX_MAP_Y) do begin
         for i := 0 to pred(MAX_MAP_X) do begin
            if Map[Emap_person,i,j] = person[k].m_number then begin
               person[k].m_x := i;
               person[k].m_y := j;
            end;
         end;
      end;

   end else begin
      randomize;
      for j := 0 to pred(MAX_MAP_Y) do begin
         for i := 0 to pred(MAX_MAP_X) do begin
            case random(100) of
                0..80 : k := 3;
               81..92 : k := 2;
               93..96 : k := 1;
               97..99 : k := 0;
            end;
            m_map_data[j][i][Emap_tile] := k;
            if random(100) < 5 then k := 4 else k := 0;
            m_map_data[j][i][Emap_height] := random(k+1);
            m_map_data[j][i][Emap_person] := 0;
         end;
      end;
   end;
end;

destructor TMap.done;
begin
end;

function  TMap.readMap(attribute : TMapAttribute; x, y : integer) : byte;

 function  isUnvalancedMap(x, y : integer; height : byte) : boolean;
 begin
    isUnvalancedMap := FALSE;
    if (x >= 0) and (y >= 0) and (x < MAX_MAP_X) and (y < MAX_MAP_Y) then begin
       if readMap(Emap_height,x,y) <> height then isUnvalancedMap := TRUE;
    end;
 end;

var
   tile_no : byte;
begin
   if (x >= 0) and (y >= 0) and (x < MAX_MAP_X) and (y < MAX_MAP_Y) then begin
      if attribute = Emap_tile then begin
         tile_no := m_map_data[y][x][Emap_tile] * 4;
         if isUnvalancedMap(x,y-1,readMap(Emap_height,x,y)) then inc(tile_no,1);
         if isUnvalancedMap(x+1,y,readMap(Emap_height,x,y)) then inc(tile_no,2);
      end else if attribute in [Emap_height,Emap_person] then begin
         tile_no := m_map_data[y][x][attribute];
      end else tile_no := 0;
   end else tile_no := 0;
   readMap := tile_no;
end;

procedure TMap.writeMap(attribute : TMapAttribute; x, y : integer; value : byte);
begin
   if (x >= 0) and (y >= 0) and (x < MAX_MAP_X) and (y < MAX_MAP_Y) then begin
      m_map_data[y][x][attribute] := value;
   end;
end;

function  TMap.isMovableMap(_x, _y, x, y : integer) : boolean;
begin
   if (Map[Emap_height,_x,_y] + 2 >= Map[Emap_height,x,y]) and
      not (Map[Emap_person,x,y] in [1..MAX_PERSON]) then
      isMovableMap := TRUE
   else
      isMovableMap := FALSE;
end;

(* ........... *)

constructor TPerson.init(number : integer);
begin
   m_number := number;
   m_x := 0;
   m_y := 0;
   m_face := 0;
   m_face_count := 0;
end;

destructor  TPerson.done;
begin
end;

procedure   TPerson.setFace(face : integer);
begin
   m_face := face;
end;

function    TPerson.returnCharaTile : integer;
begin
   returnCharaTile := pred(m_number)*PERSON_FRAME+m_face*2+m_face_count;
end;

function    TPerson.returnX : integer;
begin
   returnX := m_x;
end;

function    TPerson.returnY : integer;
begin
   returnY := m_y;
end;

function    TPerson.returnFace : integer;
begin
   returnFace := m_face;
end;

procedure   TPerson.moveXY(x1, y1 : integer);
begin
   if y1 < 0 then setFace(2);
   if y1 > 0 then setFace(1);
   if x1 < 0 then setFace(0);
   if x1 > 0 then setFace(3);
   m_face_count := m_face_count xor 1;

   if (m_x+x1 < 0) or (m_y+y1 < 0) or (m_x+x1 >= MAX_MAP_X) or (m_y+y1 >= MAX_MAP_Y) then exit;

   map.Map[Emap_person,m_x,m_y] := 0;
   if map.isMovableMap(m_x,m_y,m_x+x1,m_y+y1) then begin
      m_x := m_x + x1;
      m_y := m_y + y1;
   end else if m_number = MAIN_CHARACTER then begin
      if map.Map[Emap_person,m_x+x1,m_y+y1] in [1..MAX_PERSON] then begin
         map.Map[Emap_person,m_x,m_y] := m_number;
         exit;
      end;
      if x1 = 0 then begin
         if y1 < 0 then begin
            if map.isMovableMap(m_x,m_y,m_x-1,m_y) and map.isMovableMap(m_x,m_y,m_x-1,m_y-1) then begin
               m_x := m_x - 1;
               if map.Map[Emap_person,m_x,m_y] in [1..MAX_PERSON] then
                  person[map.Map[Emap_person,m_x,m_y]].moveXY(1,0);
            end else if map.isMovableMap(m_x,m_y,m_x+1,m_y) and map.isMovableMap(m_x,m_y,m_x+1,m_y-1) then begin
               m_x := m_x + 1;
               if map.Map[Emap_person,m_x,m_y] in [1..MAX_PERSON] then
                  person[map.Map[Emap_person,m_x,m_y]].moveXY(-1,0);
            end;
         end else begin
            if map.isMovableMap(m_x,m_y,m_x+1,m_y) and map.isMovableMap(m_x,m_y,m_x+1,m_y+1) then begin
               m_x := m_x + 1;
               if map.Map[Emap_person,m_x,m_y] in [1..MAX_PERSON] then
                  person[map.Map[Emap_person,m_x,m_y]].moveXY(-1,0);
            end else if map.isMovableMap(m_x,m_y,m_x-1,m_y) and map.isMovableMap(m_x,m_y,m_x-1,m_y+1) then begin
               m_x := m_x - 1;
               if map.Map[Emap_person,m_x,m_y] in [1..MAX_PERSON] then
                  person[map.Map[Emap_person,m_x,m_y]].moveXY(1,0);
            end;
         end;
      end else begin
         if x1 < 0 then begin
            if map.isMovableMap(m_x,m_y,m_x,m_y-1) and map.isMovableMap(m_x,m_y,m_x-1,m_y-1) then begin
               m_y := m_y - 1;
               if map.Map[Emap_person,m_x,m_y] in [1..MAX_PERSON] then
                  person[map.Map[Emap_person,m_x,m_y]].moveXY(0,1);
            end else if map.isMovableMap(m_x,m_y,m_x,m_y+1) and map.isMovableMap(m_x,m_y,m_x-1,m_y+1) then begin
               m_y := m_y + 1;
               if map.Map[Emap_person,m_x,m_y] in [1..MAX_PERSON] then
                  person[map.Map[Emap_person,m_x,m_y]].moveXY(0,-1);
            end;
         end else begin
            if map.isMovableMap(m_x,m_y,m_x,m_y+1) and map.isMovableMap(m_x,m_y,m_x+1,m_y+1) then begin
               m_y := m_y + 1;
               if map.Map[Emap_person,m_x,m_y] in [1..MAX_PERSON] then
                  person[map.Map[Emap_person,m_x,m_y]].moveXY(0,-1);
            end else if map.isMovableMap(m_x,m_y,m_x,m_y-1) and map.isMovableMap(m_x,m_y,m_x+1,m_y-1) then begin
               m_y := m_y - 1;
               if map.Map[Emap_person,m_x,m_y] in [1..MAX_PERSON] then
                  person[map.Map[Emap_person,m_x,m_y]].moveXY(0,1);
            end;
         end;
      end;
   end;
   map.Map[Emap_person,m_x,m_y] := m_number;
end;

procedure TPerson.warpXY(x, y : integer);
begin
   if (x < 0) or (y < 0) or (x >= MAX_MAP_X) or (y >= MAX_MAP_Y) then exit;
   if (m_x >= 0) and (m_y >= 0) and (m_x < MAX_MAP_X) and (m_y < MAX_MAP_Y) then begin
      if map.Map[Emap_person,m_x,m_y] = 0 then begin
         map.Map[Emap_person,m_x,m_y] := 0;
      end;
   end;
   m_x := x; m_y := y;
   map.Map[Emap_person,m_x,m_y] := m_number;
end;

procedure TPerson.doAction;
var
   x1, y1   : integer;
   key_code : word;
begin
   if m_number = MAIN_CHARACTER then begin
      if not KeyBuffer.pressedKey then exit;
      key_code := KeyBuffer.Key;
      case key_code of
         VK_UP    : moveXY(1,0);
         VK_DOWN  : moveXY(-1,0);
         VK_LEFT  : moveXY(0,-1);
         VK_RIGHT : moveXY(0,1);
         VK_ESCAPE,
         VK_F12   : Form1.Close;
         VK_PRIOR :
         begin
            MAIN_CHARACTER := MAIN_CHARACTER + 1;
            if MAIN_CHARACTER > 3 then MAIN_CHARACTER := 1;
         end;
      end;
   end else begin
      if random(10) = 0 then begin
         x1 := random(2)*2-1;
         y1 := random(2)*2-1;
         if (x1 <> 0) and (y1 <> 0) then begin
            if random(2) = 0 then x1 := 0 else y1 := 0;
         end;
         moveXY(x1,y1);
      end;
   end;
end;

(* ............ *)

procedure initializeGame;
var
   i, j, k : integer;
begin
   for i := 1 to MAX_PERSON do begin
      person[i].init(i);
   end;

   map.init;

   for j := 0 to 5 do begin
      for i := 0 to 3 do begin
         tile[j*4+i] := Bounds(i*60,succ(45*j)+1,60,45);
      end;
   end;

   for k := 0 to pred(MAX_PERSON) do begin
      for j := 0 to 3 do begin
         for i := 0 to 1 do begin
            chara[8*k+(j*2+i)] := Bounds(251+110*k+i*55,1+71*j,54,70);
         end;
      end;
   end;

   KeyBuffer.init;

end;

procedure drawScroll;
const
   X_BOUNDARY = (MAX_X_LINE-TILE_X_HALF) - (MAX_X_LINE-TILE_X_HALF) mod TILE_X_HALF;
var
   ARect                     : TRect;
   x, y, z                   : integer;
   _number, _height, _person : byte;
   x_axis,  y_axis,  x_wide  : integer;
   map_x, map_y              : integer;
   DC                        : HDC;
begin

   map_x := person[MAIN_CHARACTER].returnX;
   map_y := person[MAIN_CHARACTER].returnY;

   inc(map_x,3); inc(map_y,7);

   x_wide := 0;
   for y := - (MAP_SQ_WIDE) to (MAP_SQ_WIDE) do begin
      x_axis := (y + x_wide)*TILE_X_HALF + MAP_X_CENTER;
      y_axis := (y - x_wide)*TILE_Y_HALF + MAP_Y_CENTER;
      for x := x_wide downto -x_wide do begin
         if x_axis < MAX_X_LINE-TILE_X_SIZE then begin
            _number := map.Map[Emap_tile  ,map_x+x,map_y+y];
            _height := map.Map[Emap_height,map_x+x,map_y+y];
            _person := map.Map[Emap_person,map_x+x,map_y+y];
            for z := 0 to _height do begin
               if x_axis < X_BOUNDARY then begin
                  ARect := tile[_number];
                  with Form1 do repeat
                  until MakeItSo(BackBuffer.BltFast(x_axis,y_axis-z*13,Image[1],ARect,DDBLTFAST_SRCCOLORKEY));
               end;
            end;
            if (_person > 0) and (_person <= MAX_PERSON) then begin
               ARect := chara[person[_person].returnCharaTile];
               with Form1 do repeat
               until MakeItSo(BackBuffer.BltFast(x_axis,y_axis-_height*13-45,Image[1],ARect,DDBLTFAST_SRCCOLORKEY));
            end;
         end;

         dec(x_axis,TILE_X_HALF);
         inc(y_axis,TILE_Y_HALF);
      end;
      if y < 0 then inc(x_wide) else dec(x_wide);
   end;

   if Form1.BackBuffer.GetDC(DC) = DD_OK then begin
      with Form1.BackCanvas do try
         Handle := DC;
         Brush.Color := clBlack;
         x := 0;
         y := 450;
         FillRect(bounds(0,0,640,32));
         FillRect(bounds(0,y,640,480-y));
         FillRect(bounds(0,0,TILE_X_HALF,480));
         FillRect(bounds(X_BOUNDARY,0,640-X_BOUNDARY,480));
      finally
         Handle := 0;
         Form1.BackBuffer.ReleaseDC(DC);
      end;
   end;
end;

end.
